/*  Copyright (c) 2024 by soruce <soruce ~ AT ~ soruce.net>

    This file is part of the Arduino PimientoClockLib Library.

    The Arduino PimientoClockLib Library is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The Arduino PimientoClockLib Library is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Arduino PimientoClockLib Library.  If not, see
    <http://www.gnu.org/licenses/>.
*/

#ifndef PIMIENTO_CLOCK_LIB_H
#define PIMIENTO_CLOCK_LIB_H

#include "PimientoDisplay.h"
#include "PimientoButtons.h"

#include <uRTCLib.h>

/*
Arduino pins

DRIVER_SCLK     D13 - pin 16
DRIVER_SIN      D11 - pin 14
DRIVER_BLANK    D10 - pin 13
DRIVER_XLAT     D9  - pin 12 
DRIVER_GSCLK    D3  - pin 6

RTC_SDA         A4  - pin 23
RTC_SCL         A5  - pin 24

*/

const uint8_t MAIN_COLOR = 32;

const uint8_t BUTTON_LEFT = 0;
const uint8_t BUTTON_CENTER = 1;
const uint8_t BUTTON_RIGHT = 2;

const uint8_t EventSystemUpdate = 0;
const uint8_t EventLeftButtonReleased = 1;
const uint8_t EventRightButtonReleased = 2;
const uint8_t EventCenterButtonReleased = 3;
const uint8_t EventTimeout = 4;
const uint8_t EventLongPressCenterButton = 5;

typedef void (*Callback)(void);
typedef void (*EventCallback)(uint8_t);

class PimientoClockLib {

  public:

    /**
      Initializes relevant parts of the library logic; to be called once in the setup() of the main arduino program.
    **/
    void init(EventCallback eventCallback);

    /**
      The loop method should be called inside the main loop() in order to execute display logic, process button events and process state machine updates.
    **/
    void loop();

    // ---

    /** Runs the specified parameterless function <Callback> once.
        The <state> parameter is used to maintain state.
    **/
    static void once(uint8_t &state, Callback c);

    /** Runs the specified parameterless function <Callback> every <delay> milliseconds.
        The <counter> parameter is used for book-keeping and is updated on every call.
    **/
    static void everyMillis(const uint32_t delay, uint32_t &counter, Callback c);

    /** Runs the specified parameterless function <Callback> every <delay> frames.
        The <counter> parameter is used for book-keeping and is updated on every call.
    **/
    static void everyFrames(const uint32_t delay, uint32_t &counter, Callback c);

    /**
      Waits for the moment that the next display row needs to be updated and performs the update.
    **/
    static void waitAndUpdateActiveRow();

    // buzzer
    static void beep(uint16_t frequency, uint32_t duration);

  private:

    /** The callback method that will be called whenever any event has occurred. */
    EventCallback _eventCallback = NULL;

    static const uint8_t BUTTON_LEFT_PIN = A2;
    static const uint8_t BUTTON_CENTER_PIN = A1;
    static const uint8_t BUTTON_RIGHT_PIN = A0;

    void initRTC();
    void debugRTC();
    void processButtonEvents();
    void showDisplayDebugInfo();
};

extern uRTCLib rtc;
extern PimientoDisplay display;
extern PimientoButtons buttons;
extern PimientoClockLib pimiento;


#endif /* PIMIENTO_CLOCK_LIB_H */
