/*  Copyright (c) 2024 by soruce <soruce ~ AT ~ soruce.net>

    This file is part of the Arduino PimientoClockLib Library.

    The Arduino PimientoClockLib Library is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The Arduino PimientoClockLib Library is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Arduino PimientoClockLib Library.  If not, see
    <http://www.gnu.org/licenses/>.
*/

#include <Arduino.h>
#include <Streaming.h>

#include "IFont.h"
#include "Font1.h"
#include "Font2.h"
#include "PimientoDisplay.h"

// preinstantiated display variables
PimientoDisplay display;

Font1 font1;
Font2 font2;
IFont *fonts[2] = { &font1, &font2 };
IFont *currentFont = fonts[0];

void PimientoDisplay::init(uint16_t desiredFps) {

  // initialize pins
  pinMode(DRIVER_BLANK_PIN, OUTPUT);
  pinMode(SHIFT_LATCH_PIN, OUTPUT);
  pinMode(SHIFT_DATA_PIN, OUTPUT);
  pinMode(SHIFT_CLOCK, OUTPUT);

  // initialize display driver
  Tlc.init();

  // clear framebuffer
  clear();

  // initialize blink logic
  setBlinkSpeed(700, 300);

  // set screen refresh rate
  setRefreshRate(desiredFps);
}

void PimientoDisplay::setBlinkSpeed(uint16_t pBlinkOnDelayMillis, uint16_t pBlinkOffDelayMillis) {
  blinkOnDelayMillis = pBlinkOnDelayMillis;
  blinkOffDelayMillis = pBlinkOffDelayMillis;
}

void PimientoDisplay::updateBlink() {

  if (blinkOnDelayMillis != 0 && blinkOffDelayMillis != 0) {

    uint16_t delayMillis = blinkOn ? blinkOnDelayMillis : blinkOffDelayMillis;
    uint16_t m = millis();

    if (m - blinkCounter >= delayMillis) {
      blinkOn = !blinkOn;
      blinkCounter = m;
    }
  }
}

void PimientoDisplay::setEnabled(bool b) {
  digitalWrite(DRIVER_BLANK_PIN, b ? LOW : HIGH);
}

void PimientoDisplay::clear() {
  for (int y = 0; y < DISPLAY_HEIGHT; y++) {
    for (int x = 0; x < DISPLAY_WIDTH; x++) {
      setPixel(x, y, 0);
    }
  }
}

void PimientoDisplay::setRefreshRate(uint16_t desiredFps) {
  refreshDelayMillis = (1000 / desiredFps) / DISPLAY_HEIGHT;
}

void PimientoDisplay::setPixel(int x, int y, uint8_t color) {

  uint8_t blink = (color & 128);
  uint8_t col = (color & 127);

  col = (col * _brightness) >> 3;

  if (x >= 0 && x < DISPLAY_WIDTH && y >= 0 && y < DISPLAY_HEIGHT) {
    frameBuffer[x][y] = (col | blink);
  }
}

void PimientoDisplay::scrollFrameBufferDown() {

  uint8_t savedRow[DISPLAY_WIDTH];

  // make a copy of the last row
  for (int x = 0; x < DISPLAY_WIDTH; x++) {
    savedRow[x] = frameBuffer[x][DISPLAY_HEIGHT - 1];
  }

  for (int y = DISPLAY_HEIGHT - 1; y > 0; y--) {
    moveRowDown(y - 1, y);
  }

  // copy saved row to row 0
  for (int x = 0; x < DISPLAY_WIDTH; x++) {
    frameBuffer[x][0] = savedRow[x];
  }
}

void PimientoDisplay::scrollFrameBufferRight() {
  uint8_t savedColumn[DISPLAY_HEIGHT];

  // make a copy of the last column
  for (int y = 0; y < DISPLAY_HEIGHT; y++) {
    savedColumn[y] = frameBuffer[DISPLAY_WIDTH - 1][y];
  }

  for (int x = DISPLAY_WIDTH - 1; x > 0; x--) {
    moveColumnRight(x - 1, x);
  }

  // copy saved column to column 0
  for (int y = 0; y < DISPLAY_HEIGHT; y++) {
    frameBuffer[0][y] = savedColumn[y];
  }
}

void PimientoDisplay::moveColumnRight(int srcColumn, int destColumn) {
  for (int y = 0; y < DISPLAY_HEIGHT; y++) {
    frameBuffer[destColumn][y] = frameBuffer[srcColumn][y];
  }
}

void PimientoDisplay::moveRowDown(int srcRow, int destRow) {
  for (int x = 0; x < DISPLAY_WIDTH; x++) {
    frameBuffer[x][destRow] = frameBuffer[x][srcRow];
  }
}

void PimientoDisplay::drawRectangle(int x, int y, int width, int height, uint8_t color) {
  drawHorizontalLine(x, y, width, color);
  drawHorizontalLine(x, y + height - 1, width, color);
  drawVerticalLine(x, y, height, color);
  drawVerticalLine(x + width - 1, y, height, color);
}

void PimientoDisplay::fillRectangle(int x, int y, int width, int height, uint8_t color) {
  for (int i = 0; i < height; i++) {
    drawHorizontalLine(x, y + i, width, color);
  }
}

void PimientoDisplay::drawHorizontalLine(int x, int y, int width, uint8_t color) {
  for (int i = 0; i < width; i++) {
    setPixel(x + i, y, color);
  }
}

void PimientoDisplay::drawVerticalLine(int x, int y, int height, uint8_t color) {
  for (int i = 0; i < height; i++) {
    setPixel(x, y + i, color);
  }
}

uint32_t d1 = 0;
uint32_t d2 = 0;
uint32_t d3 = 0;
uint32_t d4 = 0;

void PimientoDisplay::update() {
  updateActiveRow();
  updateBlink();
}

void PimientoDisplay::updateActiveRow() {

  activeRow = (activeRow + 1) % DISPLAY_HEIGHT;

  uint32_t t1 = micros();

  // TODO use SPI.transfer (hardware SPI?) instead of shiftOut?
  shiftOut(SHIFT_DATA_PIN, SHIFT_CLOCK, MSBFIRST, (1 << activeRow));

  uint32_t t2 = micros();

  updateTlcData();

  uint32_t t3 = micros();

  // delay to stabilize display
  //delayMicroseconds(200);

  // pulse latch; latch signal will be used for both the 5940 and the 595 at the same time
  pulseHigh(SHIFT_LATCH_PIN);

  uint32_t t4 = micros();

  // delay to stabilize display
  //delayMicroseconds(50);

  uint32_t t5 = micros();

  rowsUpdatedCounter++;
  if (rowsUpdatedCounter % DISPLAY_HEIGHT == 0) {
    framesUpdatedCounter++;
  }

  lastRowUpdateMillis = millis();

  d1 = t2 - t1;
  d2 = t3 - t2;
  d3 = t4 - t3;
  d4 = t5 - t4;
}

void PimientoDisplay::displayDebugInfo() {

  if (debugEnabled) {
    long fps = rowsUpdatedCounter / (DISPLAY_HEIGHT * 5);
    Serial << d1 << ' ' << d2 << ' ' << d3 << ' ' << d4 << " (usec) - rowsUpdatedCounter: " << rowsUpdatedCounter << " - fps: " << fps << endl;
  }
}

void PimientoDisplay::updateTlcData() {

  for (int x = 0; x < DISPLAY_WIDTH; x++) {
    uint8_t value = frameBuffer[x][activeRow];
    uint8_t blinkEnabled = (value >> 7);
    if (blinkEnabled != 0 && !blinkOn) {
      Tlc.set(x, 0);
    } else {
      Tlc.set(x, 32 * (value & 127));
    }
  }

  Tlc.update();
  // while (Tlc.update())
  //   ;

  // while (tlc_needXLAT)
  //   ;
}

void PimientoDisplay::pulseHigh(int pin) {
  digitalWrite(pin, LOW);
  digitalWrite(pin, HIGH);
  digitalWrite(pin, LOW);
}

bool PimientoDisplay::needsActiveRowUpdate() {
  return (millis() > (lastRowUpdateMillis + refreshDelayMillis));
}

void PimientoDisplay::debugFrameBuffer() {
  for (int y = 0; y < DISPLAY_HEIGHT; y++) {
    for (int x = 0; x < DISPLAY_WIDTH; x++) {
      uint8_t value = frameBuffer[x][y];
      Serial.print(value);
      Serial.print(" ");
    }

    if (y == activeRow) {
      Serial.print(" <--");
    }

    Serial.println();
  }
}

uint8_t PimientoDisplay::drawChar(int xpos, int ypos, char c, uint8_t color) {

  int drawX = xpos;

  for (int x = 0; x < currentFont->getCharWidth(); x++) {
    uint8_t colData = currentFont->getColumnData(c, x);
    if (colData != 0) {
      for (int y = 0; y < currentFont->getCharHeight(); y++) {
        bool draw = isBitSet(colData, y);
        if (draw) {
          setPixel(drawX, ypos + y, draw ? color : 0);
        }
      }
      drawX++;
    }
  }

  return (drawX - xpos);
}

void PimientoDisplay::drawString(int xpos, int ypos, char str[], uint8_t color) {

  int x = xpos;

  for (int i = 0; i < strlen(str); i++) {
    char c = str[i];

    uint8_t columnsDrawn = drawChar(x, ypos, c, color);
    x += (columnsDrawn + currentFont->getSpacingBetweenChars());
  }
}

void PimientoDisplay::drawInt(int xpos, int ypos, int number, uint8_t color) {

  char buffer[32];
  itoa(number, buffer, 10);
  drawString(xpos, ypos, buffer, color);
}

bool PimientoDisplay::isBitSet(uint8_t value, int bitIndex) {
  return (value & (1 << bitIndex)) ? true : false;
}

void PimientoDisplay::resetCursor() {
  cursorX = 0;
  cursorY = 0;
}

void PimientoDisplay::setCursor(int x, int y) {
  cursorX = x;
  cursorY = y;
}

void PimientoDisplay::moveCursor(int x, int y) {
  cursorX += x;
  cursorY += y;
}

uint8_t PimientoDisplay::printChar(char c, uint8_t color) {
  uint8_t columnsDrawn = drawChar(cursorX, cursorY, c, color);
  cursorX += (columnsDrawn + currentFont->getSpacingBetweenChars());
}

void PimientoDisplay::printString(char str[], uint8_t color) {
  for (int i = 0; i < strlen(str); i++) {
    printChar(str[i], color);
  }
}

void PimientoDisplay::printInt(int number, uint8_t color) {
  char buffer[33];
  itoa(number, buffer, 10);
  printString(buffer, color);
}

void PimientoDisplay::printInt(int number, int numberOfDigits, uint8_t color) {

  char buffer[33];
  itoa(number, buffer, 10);

  int leadingZeroes = numberOfDigits - strlen(buffer);

  if (leadingZeroes > 0) {
    for (int i = 0; i < leadingZeroes; i++) {
      printChar('0', color);
    }
  }

  printString(buffer, color);
}

void PimientoDisplay::setFont(uint8_t fontIndex) {
  currentFont = fonts[fontIndex];
}

void PimientoDisplay::setBrightness(uint8_t brightness) {
  Serial << "_brightness: " << _brightness << endl;
  if (brightness >= 0 && brightness <= MAX_BRIGHTNESS) {
    _brightness = brightness;    
  }
  Serial << "_brightness: " << _brightness << endl;
}

uint8_t PimientoDisplay::getBrightness() {
  return _brightness;
}
